"use strict";

const remarkSchemaCustimization = require(`gatsby-transformer-remark/create-schema-customization`);

const widgetTypes = require('./widget-type-defs');

module.exports = (props, options) => {
  const {
    actions,
    schema,
    reporter
  } = props;
  const {
    config,
    widgetTypeDefs = {},
    types = {},
    remarkOptions = {
      plugins: []
    }
  } = options || {};
  const {
    createTypes,
    debug
  } = actions;
  const debugInfo = debug ? reporter.info : () => {
    return;
  };
  /*
    Setup the remark schema custimizations
  */

  remarkSchemaCustimization(props, remarkOptions);
  /*
    widgetType is an object with widgets as the key
    resolves the widget key name to a graphql type
  */

  const widgetType = Object.assign({}, widgetTypes, {}, widgetTypeDefs);

  const getType = field => {
    const {
      widget,
      required = true
    } = field;
    let type = widgetType[widget]; // default

    if (typeof type === 'object') {
      type = type.resolver ? type.resolver(field, type.default) : type.default;
      debugInfo(`${widget} Field: ${JSON.stringify(field, null, 2)}`);
    }

    return type ? `${type}${required ? '!' : ''}` : type;
  };
  /*
    Get mapped name from options types or default
  */


  const mappedName = name => {
    if (!types) return name;
    return types[name] ? types[name].type || name : name;
  };
  /*
    Get options types is node
  */


  const mappedIsNode = name => {
    if (!types) return false;
    return Boolean(types[name] && types[name].isNode);
  };
  /*
    Get options types is node
  */


  const isInferredWidget = widget => Boolean([undefined, 'hidden'].indexOf(widget));

  const mappedIsInferred = (name, widget) => {
    if (!types) return false;
    return Boolean(types[name] && types[name].infer || isInferredWidget(widget));
  };
  /*
    Get typeDefs array for a collection
    if a type has a field of sub type (object), traverse recursively
    Returns an array of typeDefs with subtypes if any
  */


  const getTypeDefs = ({
    name,
    // type name
    fields,
    // config fields to map to graphql types
    isNode,
    // is this a node type?
    infer // should gatsby infer more fields?

  }) => {
    let typeDefs = [];
    const typeFields = {};
    fields.forEach(field => {
      if (field.collection && field.valueField) {
        // Relation widget field. Write a resolver for relation field.
        const relationType = `${mappedName(field.collection)}`;
        debugInfo(`Creating Resolver ${name}.${field.name} of ${Boolean(field.multiple) ? `[${relationType}]` : `${relationType}`}`);
        typeFields[field.name] = {
          type: Boolean(field.multiple) ? `[${relationType}]` : `${relationType}`,

          resolve(parent, args, context) {
            const getValue = (obj, arr) => {
              const field = arr.shift();
              if (typeof field === 'undefined' || obj[field] === undefined) return obj;
              const subObj = obj[field];
              return getValue(subObj, [...arr]);
            };

            const isMultiple = Boolean(field.multiple);
            const valueItem = parent[`${field.name}`]; // A value field can have dot '.' notation, so later we split and getValue

            const valueField = `${field.valueField}`; // get all the nodes for the relation collections type

            const nodes = context.nodeModel.getAllNodes({
              type: `${relationType}`
            });

            if (isMultiple) {
              // Returns an array of the relation type
              return nodes.filter(node => {
                const nodeData = node.frontmatter || node;
                const nodeItem = getValue(nodeData, valueField.split('.'));
                return Boolean(valueItem.indexOf(nodeItem) + 1);
              });
            } else {
              // Returns the first single item found
              return nodes.find(node => {
                const nodeData = node.frontmatter || node;
                const nodeValue = getValue(nodeData, valueField.split('.'));
                return nodeValue === valueItem;
              });
            }
          }

        };
        return;
      }

      const type = getType(field);
      if (!type) return; // ignore this field if widget type not mapped

      if (type.startsWith('SubTypeArrayOfString')) {
        typeFields[field.name] = [`String${field.required ? '!' : ''}`];
      } else if (type.startsWith('SubType')) {
        // Object Field sub type
        const subName = `${name}_${field.name}`;
        const subType = type.startsWith('SubTypeArrayOfObject') ? [`${mappedName(subName)}!`] : `${mappedName(subName)}!`;
        typeFields[field.name] = mappedIsNode(subName) ? {
          type: subType,
          link: {} // Linked SubType (object Node)

        } : subType;
        const subDefs = getTypeDefs({
          name: subName,
          fields: field.fields,
          isNode: mappedIsNode(subName),
          infer: mappedIsInferred(subName, field.widget)
        });
        typeDefs = [...typeDefs, ...subDefs];
      } else {
        // Primitive graphql type
        typeFields[field.name] = type;
      }
    });
    const typeDef = {
      name: `${mappedName(name)}`,
      fields: typeFields,
      extensions: {
        infer: Boolean(infer)
      }
    };
    if (isNode) typeDef['interfaces'] = [`Node`];
    typeDefs.push(schema.buildObjectType(typeDef));
    return typeDefs;
  };
  /*
    typeDefs will contain all collection types (nodes)
  */


  try {
    let typeDefs = [];

    if (config && config.collections) {
      config.collections.forEach(collection => {
        if (collection.folder) {
          // Check if the type is markdown on this collection and wrap:
          // rawMarkdownBody, frontmatter, and fileAbsolutePath fields
          // frontmatter will wrap collection fields from markdown
          const isMarkdown = !collection.extension && !collection.format || Boolean([undefined, 'md', 'markdown', 'mdx'].indexOf(collection.extension) > 0) || Boolean([undefined, 'frontmatter'].indexOf(collection.format) > 0);

          if (isMarkdown) {
            // Hoist body field up from frontmatter fields
            const frontmatterFields = collection.fields.filter(field => field.name !== 'body');
            collection.fields = [{
              label: 'name',
              name: 'name',
              widget: 'string'
            }, {
              label: 'slug',
              name: 'slug',
              widget: 'string'
            }, {
              label: 'frontmatter',
              name: 'frontmatter',
              widget: 'object',
              fields: frontmatterFields
            }, {
              label: 'rawBody',
              name: 'rawBody',
              widget: 'string'
            }, {
              label: 'fileAbsolutePath',
              name: 'fileAbsolutePath',
              widget: 'string'
            }];
          }

          const defs = getTypeDefs({
            name: collection.name,
            fields: collection.fields,
            isNode: true,
            infer: mappedIsInferred(collection.name)
          });
          typeDefs = [...typeDefs, ...defs];
        } else {
          // Files collection (node with subTypes)
          const typeFields = {};
          collection.files.forEach(file => {
            const name = `${collection.name}_${file.name}`;
            const defs = getTypeDefs({
              name,
              fields: file.fields,
              isNode: true,
              // file types are always nodes
              infer: mappedIsInferred(name)
            });
            typeDefs = [...typeDefs, ...defs];
          });
        }
      });
    }

    createTypes(typeDefs);
    debugInfo(`typeDefs: ${JSON.stringify(typeDefs, null, 2)}`);
  } catch (error) {
    reporter.panic('[gatsby-source-netlify-cms] Error buiding types', error);
  }
};